import csv
import itertools
import logging
import pysam
import sys

from collections import defaultdict
from ..utils import NGSFile

logger = logging.getLogger('readlen')


def main(ngsfiles,
         outfile=sys.stdout,
         delimiter="\t",
         n_samples=100000,
         majority_vote_cutoff=0.7):

    writer = csv.DictWriter(outfile,
                            fieldnames=[
                                "File", "Evidence", "MajorityPctDetected",
                                "ConsensusReadLength"
                            ],
                            delimiter=delimiter)
    writer.writeheader()
    outfile.flush()

    if n_samples < 1:
      n_samples = None

    for ngsfilepath in ngsfiles:
        read_lengths = defaultdict(int)
        try:
            ngsfile = NGSFile(ngsfilepath)
        except:
            result = {
                "File": ngsfilepath,
                "Evidence": "File not found.",
                "MajorityPctDetected": "N/A",
                "ConsensusReadLength": "N/A"
            }

            writer.writerow(result)
            outfile.flush()
            continue

        # accumulate read lengths
        total_reads_sampled = 0
        for read in itertools.islice(ngsfile, n_samples):
            total_reads_sampled += 1
            read_lengths[len(read['query'])] += 1

        read_length_keys_sorted = sorted([int(k) for k in read_lengths.keys()],
                                         reverse=True)
        putative_max_readlen = read_length_keys_sorted[0]

        # note that simply picking the read length with the highest amount of evidence
        # doesn't make sense things like adapter trimming might shorten the read length,
        # but the read length should never grow past the maximum value.

        # if not, cannot determine, return -1
        pct = read_lengths[putative_max_readlen] / total_reads_sampled
        logger.info("Max read length percentage: {}".format(pct))
        majority_readlen = putative_max_readlen if pct > majority_vote_cutoff else -1

        result = {
            "File":
            ngsfilepath,
            "Evidence":
            ";".join([
                "{}={}".format(k, read_lengths[k])
                for k in read_length_keys_sorted
            ]),
            "MajorityPctDetected":
            round(pct, 4),
            "ConsensusReadLength":
            majority_readlen
        }

        writer.writerow(result)
        outfile.flush()



import csv
import itertools
import pysam
import sys

import logging
from collections import defaultdict

from ..utils import NGSFile, NGSFileType

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

SANGER_SET = set([i for i in range(33, 127)])
ILLUMINA_1_0_SET = set([i for i in range(59, 127)])
ILLUMINA_1_3_SET = set([i for i in range(64, 127)])

def main(ngsfiles,
         outfile=sys.stdout,
         delimiter="\t",
         n_samples=1000000):
    
    writer = csv.DictWriter(outfile,
                            fieldnames=[
                                "File", "Evidence", "ProbableEncoding"
                            ],
                            delimiter=delimiter)
    writer.writeheader()
    outfile.flush()

    if n_samples < 1:
      n_samples = None

    for ngsfilepath in ngsfiles:
        read_lengths = defaultdict(int)
        try:
            ngsfile = NGSFile(ngsfilepath, store_qualities=True)
        except FileNotFoundError:
            result = {
                "File": ngsfilepath,
                "Evidence": "File not found.",
                "ProbableEncoding": "N/A"
            }
            writer.writerow(result)
            outfile.flush()
            continue

        if ngsfile.filetype != NGSFileType.FASTQ:
            raise RuntimeError(
                "Invalid file: {}. `encoding` currently only supports FASTQ files!"
                .format(ngsfilepath))

        score_set = set()
        for read in itertools.islice(ngsfile, n_samples):
            for char in read['quality']:
                score_set.add(ord(char))
        
        max_phred_score = chr(max(score_set))
        min_phred_score = chr(min(score_set))
        if score_set <= ILLUMINA_1_3_SET:
            result = {
                "File": ngsfilepath,
                "Evidence": "Phred range: {}-{}".format(min_phred_score, max_phred_score),
                "ProbableEncoding": "Illumina 1.3"
            }
        elif score_set <= ILLUMINA_1_0_SET:
            result = {
                "File": ngsfilepath,
                "Evidence": "Phred range: {}-{}".format(min_phred_score, max_phred_score),
                "ProbableEncoding": "Solexa/Illumina 1.0"
            }
        elif score_set <= SANGER_SET:
            result = {
                "File": ngsfilepath,
                "Evidence": "Phred range: {}-{}".format(min_phred_score, max_phred_score),
                "ProbableEncoding": "Sanger/Illumina 1.8"
            }
        else:
            result = {
                "File": ngsfilepath,
                "Evidence": "Phred values outside known encoding ranges: {}-{}".format(min_phred_score, max_phred_score),
                "ProbableEncoding": "Unknown"
            }
        
        writer.writerow(result)
        outfile.flush()



import csv
import itertools
import pysam
import re
import sys

import logging
from collections import defaultdict

from ..utils import NGSFile

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

instrument_ids = {
    "^HWI-M[0-9]{4}$": ["MiSeq"],
    "^HWUSI": ["Genome Analyzer IIx"],
    "^M[0-9]{5}$": ["MiSeq"],
    "^HWI-C[0-9]{5}$": ["HiSeq 1500"],
    "^C[0-9]{5}$": ["HiSeq 1500"],
    "^HWI-ST[0-9]{3,5}(_[0-9]{9})?$": ["HiSeq 2000"],
    "^HWI-D[0-9]{5}$": ["HiSeq 2000", "HiSeq 2500"],
    "^A[0-9]{5}$": ["NovaSeq"],
    "^D[0-9]{5}$": ["HiSeq 2500"],
    "^J[0-9]{5}$": ["HiSeq 3000"],
    "^K[0-9]{5}$": ["HiSeq 3000", "HiSeq 4000"],
    "^E[0-9]{5}$": ["HiSeq X"],
    "^N[0-9]{5}$": ["NextSeq"],
    "^NB[0-9]{6}$": ["NextSeq"],
    "^NS[0-9]{6}$": ["NextSeq"],
    "^MN[0-9]{5}$": ["MiniSeq"],
}

flowcell_ids = {
    "^C[A-Z0-9]{4}ANXX$": ["HiSeq 1500", "HiSeq 2000",
                           "HiSeq 2500"],  # High Output (8-lane) v4 flow cell
    "^C[A-Z0-9]{4}ACXX$":
    ["HiSeq 1000", "HiSeq 1500", "HiSeq 2000",
     "HiSeq 2500"],  # High Output (8-lane) v3 flow cell
    "^D[A-Z0-9]{4}ACXX$":
    ["HiSeq 1000", "HiSeq 1500", "HiSeq 2000",
     "HiSeq 2500"],  # High Output (8-lane) v3 flow cell
    "^H[A-Z0-9]{4}ADXX$": ["HiSeq 1500", "HiSeq 2000",
                           "HiSeq 2500"],  # Rapid Run (2-lane) v1 flow cell
    "^H[A-Z0-9]{4}BCXX$": ["HiSeq 1500",
                           "HiSeq 2500"],  # Rapid Run (2-lane) v2 flow cell
    "^H[A-Z0-9]{4}BCXY$": ["HiSeq 1500",
                           "HiSeq 2500"],  # Rapid Run (2-lane) v2 flow cell
    "^H[A-Z0-9]{4}BBXX$": ["HiSeq 4000"],  # (8-lane) v1 flow cell
    "^H[A-Z0-9]{4}BBXY$": ["HiSeq 4000"],  # (8-lane) v1 flow cell
    "^H[A-Z0-9]{4}CCXX$": ["HiSeq X"],  # (8-lane) flow cell
    "^H[A-Z0-9]{4}CCXY$": ["HiSeq X"],  # (8-lane) flow cell
    "^H[A-Z0-9]{4}ALXX$": ["HiSeq X"],  # (8-lane) flow cell
    "^H[A-Z0-9]{4}BGX[A-Z,0-9]$": ["NextSeq"],  # High output flow cell
    "^H[A-Z0-9]{4}AFXX$": ["NextSeq"],  # Mid output flow cell
    "^H[A-Z0-9]{5}RXX$": ["NovaSeq"],  # S1 flow cell
    "^H[A-Z0-9]{5}RXX$": ["NovaSeq"],  # SP flow cell
    "^H[A-Z0-9]{5}MXX$": ["NovaSeq"],  # S2 flow cell
    "^H[A-Z0-9]{5}SXX$": ["NovaSeq"],  # S4 flow cell
    "^A[A-Z0-9]{4}$": ["MiSeq"],  # MiSeq flow cell
    "^B[A-Z0-9]{4}$": ["MiSeq"],  # MiSeq flow cell
    "^D[A-Z0-9]{4}$": ["MiSeq"],  # MiSeq nano flow cell
    # "^D[A-Z0-9]{4}$" : ["HiSeq 2000", "HiSeq 2500"],                                 # Unknown HiSeq flow cell examined in SJ data
    "^G[A-Z0-9]{4}$": ["MiSeq"],  # MiSeq micro flow cell
}

upgrade_sets = [(set(["HiSeq 2000",
                      "HiSeq 2500"]), ["HiSeq 2000", "HiSeq 2500"])]


def derive_instrument_from_iid(iid):
    matching_instruments = set()

    for pattern in instrument_ids.keys():
        if re.search(pattern, iid):
            matching_instruments |= set(instrument_ids[pattern])

    return matching_instruments


def predict_instrument_from_iids(iids):
    possible_instruments_by_iid = set()
    detected_at_least_one_instrument = False

    for iid in iids:
        if len(possible_instruments_by_iid) == 0:
            possible_instruments_by_iid = derive_instrument_from_iid(iid)
            detected_at_least_one_instrument = True
        else:
            possible_instruments_by_iid &= derive_instrument_from_iid(iid)

    return possible_instruments_by_iid, detected_at_least_one_instrument


def derive_instrument_from_fcid(fcid):
    matching_instruments = set()
    detected_at_least_one_instrument = False

    for pattern in flowcell_ids.keys():
        if re.search(pattern, fcid):
            matching_instruments |= set(flowcell_ids[pattern])

    return matching_instruments


def predict_instrument_from_fcids(fcids):
    possible_instruments_by_fcid = set()
    detected_at_least_one_instrument = False

    for fcid in fcids:
        if len(possible_instruments_by_fcid) == 0:
            possible_instruments_by_fcid = derive_instrument_from_fcid(fcid)
            detected_at_least_one_instrument = True
        else:
            possible_instruments_by_fcid &= derive_instrument_from_fcid(fcid)

    return possible_instruments_by_fcid, detected_at_least_one_instrument


def resolve_instrument(possible_instruments_by_iid,
                       possible_instruments_by_fcid,
                       at_least_one_instrument_detected,
                       malformed_read_names_detected):
    if len(possible_instruments_by_iid) == 0 and len(
            possible_instruments_by_fcid) == 0:
        if at_least_one_instrument_detected:
            return set(
                ["multiple instruments"]
            ), "unknown confidence", "multiple instruments were detected in this sample"
        elif malformed_read_names_detected:
            return set(["unknown"]), "no confidence", "read names not in Illumina format"
        else:
            return set(["unknown"]), "no confidence", "no match"

    if len(possible_instruments_by_iid) == 0:
        confidence = "medium confidence"
        if len(possible_instruments_by_fcid) > 1:
            confidence = "low confidence"
        return possible_instruments_by_fcid, confidence, "flowcell id"

    if len(possible_instruments_by_fcid) == 0:
        confidence = "medium confidence"
        if len(possible_instruments_by_fcid) > 1:
            confidence = "low confidence"
        return possible_instruments_by_iid, confidence, "instrument id"

    overlapping_instruments = possible_instruments_by_iid & possible_instruments_by_fcid
    if len(overlapping_instruments) == 0:
        return set(
            ["conflicting evidence"]
        ), "high confidence", "Case needs triaging: {} by iid, {} by fcid".format(
            " or ".join(possible_instruments_by_iid),
            " or ".join(possible_instruments_by_fcid))
    else:
        return overlapping_instruments, "high confidence", "instrument id and flowcell id"


def main(ngsfiles, outfile=sys.stdout, delimiter="\t", n_samples=10000):
    writer = csv.DictWriter(
        outfile,
        fieldnames=["File", "Instrument", "Confidence", "Basis"],
        delimiter=delimiter)
    writer.writeheader()
    outfile.flush()

    if n_samples < 1:
      n_samples = None

    for ngsfilepath in ngsfiles:
        try:
            ngsfile = NGSFile(ngsfilepath)
        except:
            result = {
                "File": ngsfilepath,
                "Instrument": "File not found.",
                "Confidence": "N/A",
                "Basis": "N/A"
            }

            writer.writerow(result)
            outfile.flush()
            continue

        instruments = set()
        flowcells = set()
        malformed_read_names = False

        # accumulate instrument and flowcell IDs
        for read in itertools.islice(ngsfile, n_samples):
            parts = read['query_name'].split(":")
            if len(parts) != 7: # not Illumina format
                malformed_read_names = True
                continue
            iid, fcid = parts[0], parts[2]
            instruments.add(iid)
            flowcells.add(fcid)

        possible_instruments_by_iid, detected_instrument_by_iid = predict_instrument_from_iids(
            instruments)
        possible_instruments_by_fcid, detected_instrument_by_fcid = predict_instrument_from_fcids(
            flowcells)

        instruments, confidence, based_on = resolve_instrument(
            possible_instruments_by_iid, possible_instruments_by_fcid,
            detected_instrument_by_iid | detected_instrument_by_fcid,
            malformed_read_names)
        for upgrade_set in upgrade_sets:
            if instruments.issubset(upgrade_set[0]):
                instruments = upgrade_set[1]
                break

        result = {
            "File": ngsfilepath,
            "Instrument": " or ".join(instruments),
            "Confidence": confidence,
            "Basis": based_on
        }

        writer.writerow(result)
        outfile.flush()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ngsderive', 'ngsderive.commands', 'ngsderive.readers']

package_data = \
{'': ['*']}

install_requires = \
['colorlog>=4.0.2,<5.0.0',
 'pysam>=0.15.3,<0.16.0',
 'pytabix>=0.0.2,<0.0.3',
 'rstr>=2.2.6,<3.0.0']

entry_points = \
{'console_scripts': ['ngsderive = ngsderive.__main__:run']}

setup_kwargs = {
    'name': 'ngsderive',
    'version': '1.2.0',
    'description': 'Forensic analysis tool useful in backwards computing information from next-generation sequencing data.',
    'long_description': '<p align="center">\n  <h1 align="center">\n    ngsderive\n  </h1>\n\n  <p align="center">\n    <a href="https://actions-badge.atrox.dev/stjudecloud/ngsderive/goto" target="_blank">\n      <img alt="Actions: CI Status"\n          src="https://img.shields.io/endpoint.svg?url=https%3A%2F%2Factions-badge.atrox.dev%2Fstjudecloud%2Fngsderive%2Fbadge&style=flat" />\n    </a>\n    <a href="https://pypi.org/project/ngsderive/" target="_blank">\n      <img alt="PyPI"\n          src="https://img.shields.io/pypi/v/ngsderive?color=orange">\n    </a>\n    <a href="https://pypi.python.org/pypi/ngsderive/" target="_blank">\n      <img alt="PyPI: Downloads"\n          src="https://img.shields.io/pypi/dm/ngsderive?color=orange">\n    </a>\n    <a href="https://pypi.python.org/pypi/ngsderive/" target="_blank">\n      <img alt="PyPI: Downloads"\n          src="https://img.shields.io/pypi/pyversions/ngsderive?color=orange">\n    </a>\n    <a href="https://github.com/stjudecloud/ngsderive/blob/master/LICENSE.md" target="_blank">\n    <img alt="License: MIT"\n          src="https://img.shields.io/badge/License-MIT-blue.svg" />\n    </a>\n  </p>\n\n\n  <p align="center">\n    Forensic analysis tool useful in backwards computing information from next-generation sequencing data. \n    <br />\n    <a href="https://stjudecloud.github.io/ngsderive/"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    <a href="https://github.com/stjudecloud/ngsderive/issues/new?assignees=&labels=&template=feature_request.md&title=Descriptive%20Title&labels=enhancement">Request Feature</a>\n    ·\n    <a href="https://github.com/stjudecloud/ngsderive/issues/new?assignees=&labels=&template=bug_report.md&title=Descriptive%20Title&labels=bug">Report Bug</a>\n    ·\n    ⭐ Consider starring the repo! ⭐\n    <br />\n  </p>\n</p>\n\n> Notice: `ngsderive` is a forensic analysis tool useful in backwards computing information \n> from next-generation sequencing data. Notably, results are provided as a \'best guess\' — \n> the tool does not claim 100% accuracy and results should be considered with that understanding.\n\n## 🎨 Features\n\nThe following attributes can be guessed using ngsderive:\n\n* <b>Illumina Instrument.</b> Infer which Illumina instrument was used to generate the data by matching against known instrument and flowcell naming patterns. Each guess comes with a confidence score. \n* <b>RNA-Seq Strandedness.</b> Infer from the data whether RNA-Seq data was generated using a Stranded-Forward, Stranded-Reverse, or Unstranded protocol.\n* <b>Pre-trimmed Read Length.</b> Compute the distribution of read lengths in the file and attempt to guess what the original read length of the experiment was.\n\n## 📚 Getting Started\n\n### Installation\n\nYou can install ngsderive using the Python Package Index ([PyPI](https://pypi.org/)).\n\n```bash\npip install ngsderive\n```\n\n## 🖥️ Development\n\nIf you are interested in contributing to the code, please first review\nour [CONTRIBUTING.md][contributing-md] document. \n\nTo bootstrap a development environment, please use the following commands.\n\n```bash\n# Clone the repository\ngit clone git@github.com:stjudecloud/ngsderive.git\ncd ngsderive\n\n# Install the project using poetry\npoetry install\n```\n\n## 🚧️ Tests\n\nngsderive provides a (currently patchy) set of tests — both unit and end-to-end.\n\n```bash\npy.test\n```\n\n## 🤝 Contributing\n\nContributions, issues and feature requests are welcome!<br />Feel free to check [issues page](https://github.com/stjudecloud/ngsderive/issues). You can also take a look at the [contributing guide][contributing-md].\n\n## 📝 License\n\nThis project is licensed as follows:\n\n* All code related to the `instrument` subcommand is licensed under the [AGPL\n  v2.0][agpl-v2]. This is not due any strict requirement, but out of deference\n  to some [code][10x-inspiration] I drew inspiration from (and copied patterns\n  from), the decision was made to license this code consistently.\n* The rest of the project is licensed under the MIT License - see the\n  [LICENSE.md](LICENSE.md) file for details.\n\nCopyright © 2020 [St. Jude Cloud Team](https://github.com/stjudecloud).<br />\n\n[10x-inspiration]: https://github.com/10XGenomics/supernova/blob/master/tenkit/lib/python/tenkit/illumina_instrument.py\n[agpl-v2]: http://www.affero.org/agpl2.html\n[contributing-md]: https://github.com/stjudecloud/ngsderive/blob/master/CONTRIBUTING.md\n[license-md]: https://github.com/stjudecloud/ngsderive/blob/master/LICENSE.md\n',
    'author': 'Clay McLeod',
    'author_email': 'Clay.McLeod@STJUDE.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/claymcleod/ngsderive',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
